/*
 *	BOARD.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */

/*
 *  Simple Bitboard move generator by Michael Sherwin
 */

#include <stdlib.h>
#include "defs.h"
#include "data.h"
#include "protos.h"


/* init_board() sets the board to the initial game state. */

void init_board()
{
	int i;

      light_squares = 0;
      dark_squares = 0;

	for (i = 0; i < 64; ++i) {
		color[i] = init_color[i];
		piece[i] = init_piece[i];
            if (color[i] == LIGHT)
                  light_squares |= set_mask[i];
            if (color[i] == DARK)
                  dark_squares |= set_mask[i];
	}
	side = LIGHT;
	xside = DARK;
	castle = 15;
	ep = -1;
	fifty = 0;
	ply = 0;
	hply = 0;
	set_hash();  /* init_hash() must be called before this function */
	first_move[0] = 0;
}

void init_bitboard()
{
      int i, j, n;
      
      /* mask used to set or clear one bit of 64 bits. */
      for (i = 0; i < 64; ++i) {
            set_mask[i] = (u64)1 << i;
            clear_mask[i] = 0xffffffffffffffff ^ ((u64)1 << i);
      }
      set_mask[64] = 0;
      clear_mask[64] = 0xffffffffffffffff;
      above[64] = 0xffffffffffffffff;
      below[64] = 0xffffffffffffffff;

      for (i = 0; i < 64; ++i) {

            /* mask used to select all bits above bit i of 64 bits. */
            above[i] = 0;
            for (j = i + 1; j < 64; ++j)
                  above[i] |= set_mask[j];

            /* mask used to select all bits below bit i of 64 bits. */
            below[i] = 0;
            for (j = 0; j < i; ++j)
                  below[i] |= set_mask[j];

            /* empty board pawn moves from square i */
            light_pawn_moves[i] = 0;
            light_pawn_captures[i] = 0;
            dark_pawn_moves[i] = 0;
            dark_pawn_captures[i] = 0;
            if (i > 7 && i < 56) {
                  light_pawn_moves[i] = set_mask[i - 8];
                  dark_pawn_moves[i] = set_mask[i + 8];
                  if (i > 47)
                        light_pawn_moves[i] |= set_mask[i - 16];
                  if (i < 16)
                        dark_pawn_moves[i] |= set_mask[i + 16];
                  if (COL(i) != 0) {
                        light_pawn_captures[i] = set_mask[i - 9];
                        dark_pawn_captures[i] = set_mask[i + 7];
                  } 
                  if (COL(i) != 7) {
                        light_pawn_captures[i] |= set_mask[i - 7];
                        dark_pawn_captures[i] |= set_mask[i + 9];
                  } 
            }
            /* empty board knight moves from square i */
            knight_moves[i] = 0;
            for (j = 0; j < offsets[KNIGHT]; ++j) {
		           	   n = i;
			             n = mailbox[mailbox64[n] + offset[KNIGHT][j]];
			             if (n == -1)
				                 continue;
                   knight_moves[i] |= set_mask[n];
            } 

            /* empty board moves for each ray from square i. */
            for (j = 0; j < offsets[QUEEN]; ++j) {
                   *(rays[j] + i) = 0;
			             for (n = i;;) {
			                  	n = mailbox[mailbox64[n] + offset[QUEEN][j]];
				                  if (n == -1)
					                      break;
                          *(rays[j] + i) |= set_mask[n];
                  }
            }

            /* empty board king moves from square i. */
            for (j = 0; j < offsets[KING]; ++j) {
			             n = i;
			             n = mailbox[mailbox64[n] + offset[KING][j]];
			             if (n == -1)
				                 continue;
                         king_moves[i] |= set_mask[n];
            }
      }
}      


/* init_hash() initializes the random numbers used by set_hash(). */

void init_hash()
{
	int i, j, k;

	srand(0);
	for (i = 0; i < 2; ++i)
		for (j = 0; j < 6; ++j)
			for (k = 0; k < 64; ++k)
				hash_piece[i][j][k] = hash_rand();
	hash_side = hash_rand();
	for (i = 0; i < 64; ++i)
		hash_ep[i] = hash_rand();
}


/* hash_rand() XORs some shifted random numbers together to make sure
   we have good coverage of all 32 bits. (rand() returns 16-bit numbers
   on some systems.) */

int hash_rand()
{
	int i;
	int r = 0;

	for (i = 0; i < 32; ++i)
		r ^= rand() << i;
	return r;
}


/* set_hash() uses the Zobrist method of generating a unique number (hash)
   for the current chess position. Of course, there are many more chess
   positions than there are 32 bit numbers, so the numbers generated are
   not really unique, but they're unique enough for our purposes (to detect
   repetitions of the position). 
   The way it works is to XOR random numbers that correspond to features of
   the position, e.g., if there's a black knight on B8, hash is XORed with
   hash_piece[BLACK][KNIGHT][B8]. All of the pieces are XORed together,
   hash_side is XORed if it's black's move, and the en passant square is
   XORed if there is one. (A chess technicality is that one position can't
   be a repetition of another if the en passant state is different.) */

void set_hash()
{
  u64 squares;
	int i;

  squares = light_squares | dark_squares;
	hash = 0;	
  while(squares) {
          i = first_one(squares);
          squares &= clear_mask[i];
    			hash ^= hash_piece[color[i]][piece[i]][i];
  }
	if (side == DARK)
		hash ^= hash_side;
	if (ep != -1)
		hash ^= hash_ep[ep];
}


/* in_check() returns TRUE if side s is in check and FALSE
   otherwise. It just scans the board to find side s's king
   and calls attack() to see if it's being attacked. */

BOOL in_check(int s)
{
	int i;
  u64 squares;
  int (*next_one)(u64);

  if (s == LIGHT)
  {
        squares = light_squares;
        next_one = last_one;
  }
  else
  {
        squares = dark_squares;
        next_one = first_one;
  }

  while (squares) {
           i = next_one(squares);
           squares &= clear_mask[i];
	         if (piece[i] == KING)
	  	     return attack(i, s ^ 1);
  } 
	return TRUE;  /* shouldn't get here */
}


/* attack() returns TRUE if square sq is being attacked by side
   s and FALSE otherwise. */

BOOL attack(int sq, int s)
{
      
	    int i, j, n;
      u64 squares;

      if ( s == LIGHT)
      {
            squares = light_squares;
      }
      else
      {
            squares = dark_squares;
      }

      while (squares) {
            i = first_one(squares);
            squares &= clear_mask[i];
			if (piece[i] == PAWN) {
				if (s == LIGHT) {
					if (COL(i) != 0 && i - 9 == sq)
						return TRUE;
					if (COL(i) != 7 && i - 7 == sq)
						return TRUE;
				}
				else {
					if (COL(i) != 0 && i + 7 == sq)
						return TRUE;
					if (COL(i) != 7 && i + 9 == sq)
						return TRUE;
				}
			}
			else
				for (j = 0; j < offsets[piece[i]]; ++j)
					for (n = i;;) {
						n = mailbox[mailbox64[n] + offset[piece[i]][j]];
						if (n == -1)
							break;
						if (n == sq)
							return TRUE;
						if (color[n] != EMPTY)
							break;
						if (!slide[piece[i]])
							break;
					}
      }
	return FALSE;
}


/* gen() generates pseudo-legal moves for the current position.
   It scans the board to find friendly pieces and then determines
   what squares they attack. When it finds a piece/square
   combination, it calls gen_push to put the move on the "move
   stack." */

void gen()
{
      int fs;
      int ts;
      int first;
      u64 squares;
      u64 occupied;
      u64 enemy;
      u64 moves;
      u64 captures;
      int (*next_one)(u64);
      
	    first_move[ply + 1] = first_move[ply];
      if (side == LIGHT) {
            squares = light_squares;
            enemy = dark_squares;
            next_one = first_one;
      }
      else {
            squares = dark_squares;
            enemy = light_squares;
            next_one = last_one;
      }
      occupied = light_squares | dark_squares;
      while (squares) {
        fs = next_one (squares);
        squares &= clear_mask[fs];
        switch (piece[fs]) {
        case PAWN:
          if (side == LIGHT) {
              first = last_one (light_pawn_moves[fs] & occupied);
              moves = light_pawn_moves[fs] & above[first];
              captures = light_pawn_captures[fs] & (enemy | set_mask[ep]);
              while(moves) {
                ts = first_one(moves);
                moves &= clear_mask[ts];
                if(fs >= 48 && ts <= 39)
                  gen_push(fs, ts, 24);
                else
                  gen_push(fs, ts, 16);
              }  
          }
          else {
              first = first_one (dark_pawn_moves[fs] & occupied);
              moves = dark_pawn_moves[fs] & below[first];
              captures = dark_pawn_captures[fs] & (enemy | set_mask[ep]);
              while(moves) {
                ts = first_one(moves);
                moves &= clear_mask[ts];
                if(fs <= 15 && ts >= 24)
                  gen_push(fs, ts, 24);
                else
                  gen_push(fs, ts, 16);
              }  
          }
          while(captures) {
            ts = first_one(captures);
            captures &= clear_mask[ts];
            if(ts == ep)
              gen_push(fs, ts, 21);
            else
              gen_push(fs, ts, 17);
          } 
          continue;
        case KNIGHT:
          moves = knight_moves[fs] & ~occupied;
          captures = knight_moves[fs] & enemy;
          break;
        case BISHOP:
          first = first_one(ray_plus_7[fs] & occupied);
          moves = ray_plus_7[fs] & below[first];
          captures = set_mask[first];
          first = first_one(ray_plus_9[fs] & occupied);
          moves |= ray_plus_9[fs] & below[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_7[fs] & occupied);
          moves |= ray_minus_7[fs] & above[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_9[fs] & occupied);
          moves |= ray_minus_9[fs] & above[first];
          captures = (captures | set_mask[first]) & enemy;
          break;
        case ROOK:
          first = first_one(ray_plus_8[fs] & occupied);
          moves = ray_plus_8[fs] & below[first];
          captures = set_mask[first];
          first = first_one(ray_plus_1[fs] & occupied);
          moves |= ray_plus_1[fs] & below[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_8[fs] & occupied);
          moves |= ray_minus_8[fs] & above[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_1[fs] & occupied);
          moves |= ray_minus_1[fs] & above[first];
          captures = (captures | set_mask[first]) & enemy;
          break;
        case QUEEN:
          first = first_one(ray_plus_7[fs] & occupied);
          moves = ray_plus_7[fs] & below[first];
          captures = set_mask[first];
          first = first_one(ray_plus_9[fs] & occupied);
          moves |= ray_plus_9[fs] & below[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_7[fs] & occupied);
          moves |= ray_minus_7[fs] & above[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_9[fs] & occupied);
          moves |= ray_minus_9[fs] & above[first];
          captures |= set_mask[first];
          first = first_one(ray_plus_8[fs] & occupied);
          moves |= ray_plus_8[fs] & below[first];
          captures |= set_mask[first];
          first = first_one(ray_plus_1[fs] & occupied);
          moves |= ray_plus_1[fs] & below[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_8[fs] & occupied);
          moves |= ray_minus_8[fs] & above[first];
          captures |= set_mask[first];
          first = last_one(ray_minus_1[fs] & occupied);
          moves |= ray_minus_1[fs] & above[first];
          captures = (captures | set_mask[first]) & enemy;
          break;
        case KING:
          moves = king_moves[fs] & ~occupied;
          captures = king_moves[fs] & enemy;
          break;
        }
        while(captures) {
          ts = first_one(captures);
          captures &= clear_mask[ts];
          gen_push(fs, ts, 1);
        }
        while(moves) {
          ts = first_one(moves);
          moves &= clear_mask[ts];
          gen_push(fs, ts, 0);
        }
      }

	/* generate castle moves */
	if (side == LIGHT) {
		if (castle & 1)
			gen_push(E1, G1, 2);
		if (castle & 2)
			gen_push(E1, C1, 2);
	}
	else {
		if (castle & 4)
			gen_push(E8, G8, 2);
		if (castle & 8)
			gen_push(E8, C8, 2);
	}
	
}


/* gen_caps() is basically a copy of gen() that's modified to
   only generate capture and promote moves. It's used by the
   quiescence search. */

void gen_caps()
{
      int fs;
      int ts;
      int first;
      u64 squares;
      u64 occupied;
      u64 enemy;
      u64 moves;
      u64 captures;
      int (*next_one)(u64);

	    first_move[ply + 1] = first_move[ply];
      if (side == LIGHT) {
            squares = light_squares;
            enemy = dark_squares;
            next_one = first_one;
      }
      else {
            squares = dark_squares;
            enemy = light_squares;
            next_one = last_one;
      }
      occupied = light_squares | dark_squares;
      while (squares) {
        fs = next_one (squares);
        squares &= clear_mask[fs];
        switch (piece[fs]) {
        case PAWN:
          if (side == LIGHT) {
              first = last_one (light_pawn_moves[fs] & occupied);
              moves = (light_pawn_moves[fs] & above[first]) & below[8];
              captures = light_pawn_captures[fs] & (enemy | set_mask[ep]);
          }
          else {
              first = first_one (dark_pawn_moves[fs] & occupied);
              moves = (dark_pawn_moves[fs] & below[first]) & above[55];
              captures = dark_pawn_captures[fs] & (enemy | set_mask[ep]);
          }
          while(captures) {
            ts = first_one(captures);
            captures &= clear_mask[ts];
            if(ts == ep)
              gen_push(fs, ts, 21);
            else
              gen_push(fs, ts, 17);
          } 
          while(moves) {
            ts = first_one(moves);
            moves &= clear_mask[ts];
            gen_push(fs, ts, 16);
          }  
          continue;
        case KNIGHT:
          captures = knight_moves[fs] & enemy;
          break;
        case BISHOP:
          captures = (set_mask[first_one(ray_plus_7[fs] & occupied)]
                   |  set_mask[first_one(ray_plus_9[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_7[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_9[fs] & occupied)]) & enemy;
          break;
        case ROOK:
          captures = (set_mask[first_one(ray_plus_8[fs] & occupied)]
                   |  set_mask[first_one(ray_plus_1[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_8[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_1[fs] & occupied)]) & enemy;
          break;
        case QUEEN:
          captures = (set_mask[first_one(ray_plus_7[fs] & occupied)]
                   |  set_mask[first_one(ray_plus_9[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_7[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_9[fs] & occupied)]
                   |  set_mask[first_one(ray_plus_8[fs] & occupied)]
                   |  set_mask[first_one(ray_plus_1[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_8[fs] & occupied)]
                   |  set_mask[last_one(ray_minus_1[fs] & occupied)]) & enemy;
          break;
        case KING:
          captures = king_moves[fs] & enemy;
          break;
        }
        while(captures) {
          ts = first_one(captures);
          captures &= clear_mask[ts];
          gen_push(fs, ts, 1);
        }
      }
}


/* gen_push() puts a move on the move stack, unless it's a
   pawn promotion that needs to be handled by gen_promote().
   It also assigns a score to the move for alpha-beta move
   ordering. If the move is a capture, it uses MVV/LVA
   (Most Valuable Victim/Least Valuable Attacker). Otherwise,
   it uses the move's history heuristic value. Note that
   1,000,000 is added to a capture move's score, so it
   always gets ordered above a "normal" move. */

void gen_push(int from, int to, int bits)
{
	gen_t *g;
	
	if (bits & 16) {
		if (side == LIGHT) {
			if (to <= H8) {
				gen_promote(from, to, bits);
				return;
			}
		}
		else {
			if (to >= A1) {
				gen_promote(from, to, bits);
				return;
			}
		}
	}
	g = &gen_dat[first_move[ply + 1]++];
	g->m.b.from = (char)from;
	g->m.b.to = (char)to;
	g->m.b.promote = 0;
	g->m.b.bits = (char)bits;
	if (color[to] != EMPTY)
		g->score = 1000000 + (piece[to] * 10) - piece[from];
	else
		g->score = history[from][to];
}


/* gen_promote() is just like gen_push(), only it puts 4 moves
   on the move stack, one for each possible promotion piece */

void gen_promote(int from, int to, int bits)
{
	int i;
	gen_t *g;
	
	for (i = KNIGHT; i <= QUEEN; ++i) {
		g = &gen_dat[first_move[ply + 1]++];
		g->m.b.from = (char)from;
		g->m.b.to = (char)to;
		g->m.b.promote = (char)i;
		g->m.b.bits = (char)(bits | 32);
		g->score = 1000000 + (i * 10);
	}
}


/* makemove() makes a move. If the move is illegal, it
   undoes whatever it did and returns FALSE. Otherwise, it
   returns TRUE. */

BOOL makemove(move_bytes m)
{
	
	/* test to see if a castle move is legal and move the rook
	   (the king is moved with the usual move code later) */
	if (m.bits & 2) {
		int from, to;

		if (in_check(side))
			return FALSE;
		switch (m.to) {
			case 62:
				if (color[F1] != EMPTY || color[G1] != EMPTY ||
						attack(F1, xside) || attack(G1, xside))
					return FALSE;
				from = H1;
				to = F1;
                        light_squares &= clear_mask[from];
                        light_squares |= set_mask[to];
				break;
			case 58:
				if (color[B1] != EMPTY || color[C1] != EMPTY || color[D1] != EMPTY ||
						attack(C1, xside) || attack(D1, xside))
					return FALSE;
				from = A1;
				to = D1;
                        light_squares &= clear_mask[from];
                        light_squares |= set_mask[to];
				break;
			case 6:
				if (color[F8] != EMPTY || color[G8] != EMPTY ||
						attack(F8, xside) || attack(G8, xside))
					return FALSE;
				from = H8;
				to = F8;
                        dark_squares &= clear_mask[from];
                        dark_squares |= set_mask[to];
				break;
			case 2:
				if (color[B8] != EMPTY || color[C8] != EMPTY || color[D8] != EMPTY ||
						attack(C8, xside) || attack(D8, xside))
					return FALSE;
				from = A8;
				to = D8;
                        dark_squares &= clear_mask[from];
                        dark_squares |= set_mask[to];
				break;
			default:  /* shouldn't get here */
				from = -1;
				to = -1;
				break;
		}
		color[to] = color[from];
		piece[to] = piece[from];
		color[from] = EMPTY;
		piece[from] = EMPTY;
	}

	/* back up information so we can take the move back later. */
	hist_dat[hply].m.b = m;
	hist_dat[hply].capture = piece[(int)m.to];
      hist_dat[hply].castle = castle;
	hist_dat[hply].ep = ep;
	hist_dat[hply].fifty = fifty;
	hist_dat[hply].hash = hash;
	++ply;
	++hply;

	/* update the castle, en passant, and
	   fifty-move-draw variables */
	castle &= castle_mask[(int)m.from] & castle_mask[(int)m.to];
	if (m.bits & 8) {
		if (side == LIGHT)
			ep = m.to + 8;
		else
			ep = m.to - 8;
	}
	else
		ep = -1;
	if (m.bits & 17)
		fifty = 0;
	else
		++fifty;

	/* move the piece */
	color[(int)m.to] = side;
	if (m.bits & 32)
		piece[(int)m.to] = m.promote;
	else
		piece[(int)m.to] = piece[(int)m.from];
	color[(int)m.from] = EMPTY;
	piece[(int)m.from] = EMPTY;
      if (side == LIGHT) {
            light_squares &= clear_mask[(int)m.from];
            light_squares |= set_mask[(int)m.to];
            dark_squares &= clear_mask[(int)m.to];
      }
      else {
            dark_squares &= clear_mask[(int)m.from];
            dark_squares |= set_mask[(int)m.to];
            light_squares &= clear_mask[(int)m.to];
      }
	/* erase the pawn if this is an en passant move */
	if (m.bits & 4) {
		if (side == LIGHT) {
			color[m.to + 8] = EMPTY;
			piece[m.to + 8] = EMPTY;
                  dark_squares &= clear_mask[m.to + 8];
		}
		else {
			color[m.to - 8] = EMPTY;
			piece[m.to - 8] = EMPTY;
                  light_squares &= clear_mask[m.to - 8];
		}
	}

	/* switch sides and test for legality (if we can capture
	   the other guy's king, it's an illegal position and
	   we need to take the move back) */
	side ^= 1;
	xside ^= 1;
	if (in_check(xside)) {
		takeback();
		return FALSE;
	}
	set_hash();
	return TRUE;
}


/* takeback() is very similar to makemove(), only backwards :)  */

void takeback()
{
	move_bytes m;

	side ^= 1;
	xside ^= 1;
	--ply;
	--hply;
	m = hist_dat[hply].m.b;
	castle = hist_dat[hply].castle;
	ep = hist_dat[hply].ep;
	fifty = hist_dat[hply].fifty;
	hash = hist_dat[hply].hash;
	color[(int)m.from] = side;
	if (m.bits & 32)
		piece[(int)m.from] = PAWN;
	else
		piece[(int)m.from] = piece[(int)m.to];
	if (hist_dat[hply].capture == EMPTY) {
		color[(int)m.to] = EMPTY;
		piece[(int)m.to] = EMPTY;
	}
	else {
		color[(int)m.to] = xside;
		piece[(int)m.to] = hist_dat[hply].capture;
            if (side == LIGHT)
                  dark_squares |= set_mask[m.to];
            else
                  light_squares |= set_mask[m.to];
	}
      if (side == LIGHT) {
            light_squares &= clear_mask[m.to];
            light_squares |= set_mask[m.from];
      }
      else {
            dark_squares &= clear_mask[m.to];
            dark_squares |= set_mask[m.from];
      }
	if (m.bits & 2) {
		int from, to;

		switch(m.to) {
			case 62:
				from = F1;
				to = H1;
                        light_squares &= clear_mask[from];
                        light_squares |= set_mask[to];
				break;
			case 58:
				from = D1;
				to = A1;
                        light_squares &= clear_mask[from];
                        light_squares |= set_mask[to];
				break;
			case 6:
				from = F8;
				to = H8;
                        dark_squares &= clear_mask[from];
                        dark_squares |= set_mask[to];
				break;
			case 2:
				from = D8;
				to = A8;
                        dark_squares &= clear_mask[from];
                        dark_squares |= set_mask[to];
				break;
			default:  /* shouldn't get here */
				from = -1;
				to = -1;
				break;
		}
		color[to] = side;
		piece[to] = ROOK;
		color[from] = EMPTY;
		piece[from] = EMPTY;
	}
	if (m.bits & 4) {
		if (side == LIGHT) {
			color[m.to + 8] = xside;
			piece[m.to + 8] = PAWN;
                  dark_squares |= set_mask[m.to + 8];
		}
		else {
			color[m.to - 8] = xside;
			piece[m.to - 8] = PAWN;
                  light_squares |= set_mask[m.to - 8];
		}
	}
}
